
//
// MailDirDriver.cs
//
// Copyright (C) 2005 Noam Rathaus, Aviram Jenik
//
//
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//


using System;
using System.Collections;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Threading;

using Beagle.Util;

namespace Beagle.Daemon.MailDirDriver {

	[QueryableFlavor (Name="Mail", Domain=QueryDomain.Local, RequireInotify=false)]
	public class MailDirQueryable : LuceneQueryable {

		public int polling_interval_in_seconds = 60;

		public static Logger log = Logger.Get ("mail");
		private string path;

		private SortedList watched = new SortedList ();
		private MailDirCrawler crawler;

		private object lockObj = new object ();
		private ArrayList AddedUris = new ArrayList ();
		private bool queryRunning = false;

		private string folder_cache_name;
	  private Hashtable mapping;

		private long index_count;

		public MailDirQueryable () : base ("MailDirIndex")
		{
			string my_path = "Mail";
			this.path = Path.Combine (PathFinder.HomeDir, my_path);
			// FIXME: Other paths to add here? maybe KMail's DIMAP and IMAP directories?
		}

		private void Crawl ()
		{
			crawler.Crawl ();
			foreach (FileInfo file in crawler.Emails)
				IndexEmail (file);
		}

		private void CrawlHook (Scheduler.Task task)
		{
			Crawl ();
			task.Reschedule = true;
			task.TriggerTime = DateTime.Now.AddSeconds (polling_interval_in_seconds);
		}

		private void StartWorker ()
		{
			Logger.Log.Info ("Starting Mail Dir mail backend");

			Stopwatch stopwatch = new Stopwatch ();
			stopwatch.Start ();

			// Check that we have data to index
			if (! Directory.Exists (this.path)) {
				// No mails present, repoll every minute
				log.Warn ("Mail Dir mail store not found, watching for it.");
				GLib.Timeout.Add (60000, new GLib.TimeoutHandler (CheckForMailData));
				return;
			}
			else
			{
				Logger.Log.Info ("Mail Dir found, path: {0}", this.path);
			}

			// Get notification when an index or summary file changes
			if (Inotify.Enabled) {
				Inotify.Event += OnInotifyEvent;
				Watch (this.path);
			}

			Logger.Log.Debug ("Starting Mail Dir crawl");
			crawler = new MailDirCrawler (this.path);
			Crawl ();
			Logger.Log.Debug ("Mail Dir crawl finished");

			// If we don't have inotify, we have to poll the file system.  Ugh.
			if (! Inotify.Enabled) {
				Logger.Log.Info ("INotify missing, Mail Dir will be watched manaually");

				Scheduler.Task task = Scheduler.TaskFromHook (new Scheduler.TaskHook (CrawlHook));
				task.Tag = "Crawling Kmail dir to find emails";
				ThisScheduler.Add (task);
			}

			stopwatch.Stop ();
			Logger.Log.Info ("MailDir driver worker thread done in {0}",
					 stopwatch);
			Logger.Log.Info ("Saving cache");
			this.SaveCache();
		}

		public override void Start () 
		{
			Logger.Log.Info ("Starting Mail Dir backend");
			/* make sure cache is loaded */
	      if (this.mapping == null) {
    	    bool cache_loaded = this.LoadCache ();
        	if (!cache_loaded) {
				MailDirQueryable.log.Debug("Error! could not load cache");
				this.mapping = new Hashtable();
				this.SaveCache();
			}
		  }
		  
		  base.Start ();
		  ExceptionHandlingThread.Start (new ThreadStart (StartWorker));
		}

		private string FolderCacheName {
			get {
				if (this.folder_cache_name == null)
					this.folder_cache_name = "status-" + "Maildir"; // FIXME: Is this filename acceptable?
				
				return this.folder_cache_name;
			}
		}

    private bool LoadCache ()
    {
      Stream cacheStream;
      BinaryFormatter formatter;

      try {
        cacheStream = ReadAppData (this.FolderCacheName);
        formatter = new BinaryFormatter ();
        this.mapping = formatter.Deserialize (cacheStream) as Hashtable;
        cacheStream.Close ();
        MailDirQueryable.log.Debug ("Successfully loaded previous crawled data from disk: {0}", this.FolderCacheName);

        return true;
      } catch {
        this.mapping = new Hashtable ();

        return false;
      }
    }
		
		private void SaveCache ()
    {
      Stream cacheStream;
      BinaryFormatter formatter;

      cacheStream = WriteAppData (this.FolderCacheName);
      formatter = new BinaryFormatter ();
      formatter.Serialize (cacheStream, mapping);
      cacheStream.Close ();
    }


		private void Watch (string path)
		{
			DirectoryInfo root = new DirectoryInfo (path);
			if (! root.Exists)
				return;

			Queue queue = new Queue ();
			queue.Enqueue (root);

			while (queue.Count > 0) {
				DirectoryInfo dir = queue.Dequeue () as DirectoryInfo;

				if (! dir.Exists)
					continue;
				
				int wd = Inotify.Watch (dir.FullName,
						//FIXME: Does these Inotify events cover all possibilities?
													Inotify.EventType.Create
													| Inotify.EventType.Delete
													| Inotify.EventType.MovedTo
													| Inotify.EventType.MovedFrom
													);
				watched [wd] = dir.FullName;

				foreach (DirectoryInfo subdir in dir.GetDirectories ())
					queue.Enqueue (subdir);
			}
		}

		private void Ignore (string path)
		{
			Inotify.Ignore (path);
			watched.RemoveAt (watched.IndexOfValue (path));
		}

		private void OnInotifyEvent (int wd,
					     string path,
					     string subitem,
					     string srcpath,
					     Inotify.EventType type)
		{
			if (subitem == "" || ! watched.Contains (wd))
				return;

			string fullPath = Path.Combine (path, subitem);

			if ((type & Inotify.EventType.Delete) != 0)
			{
				if ((type & Inotify.EventType.IsDirectory) != 0)
				{
					log.Debug ("Delete directory event called");
					Ignore (fullPath);
				}
				else
				{
					log.Debug ("Delete file event called");
					this.Remove(new FileInfo(fullPath));
					this.mapping.Remove(fullPath);
				}
			}
			
			if ((type & Inotify.EventType.Create) != 0)
			{
				// Used by Subdirs Events
				log.Info ("Create event called");
				Watch (fullPath);
			}

			if ((type & Inotify.EventType.MovedTo) != 0)
			{
				log.Info ("MovedTo event called");
				log.Info ("Indexing new file: {0}", fullPath);
				this.IndexEmail (new FileInfo (fullPath));
			}

			if ((type & Inotify.EventType.MovedFrom) != 0)
			{
				log.Info ("MovedFrom event called");
				log.Info ("Removing index for file: {0}", fullPath);
				this.Remove(new FileInfo(fullPath));
				this.mapping.Remove(fullPath);
			}
		}

		private void Remove(FileInfo file)
		{
			Uri file_uri = UriFu.PathToFileUri(file.FullName);
      
			Scheduler.Task task;
			task = NewRemoveTask (file_uri);
			task.Priority = Scheduler.Priority.Immediate;
			ThisScheduler.Add (task);	
		}

		private bool CheckForMailData ()
		{
			if (! Directory.Exists (this.path))
				return true; // continue polling
			
			// Otherwise stop polling and start indexing
			StartWorker();
			return false;
		}

		public string Name {
			get { return "MailDir"; }
		}

		public void IndexEmail (FileInfo emailInfo)
		{
		 string filename = emailInfo.FullName;
		 if(this.mapping[filename]==null) {
		 		++this.index_count;
		  	MailDirQueryable.log.Debug("{0} not found in cache", filename);
		 		MailDirIndexableGenerator generator = new MailDirIndexableGenerator (this, emailInfo);
				Scheduler.Task task;
				
				task = NewAddTask (generator, new Scheduler.Hook (generator.Checkpoint));
				// IndexableGenerator tasks default to having priority Scheduler.Priority Generator
				ThisScheduler.Add (task);
				this.mapping[filename] = filename; // FIXME: We add to the cache without verifying whether the email got index or not... need to fix
				}
	      else {
       		MailDirQueryable.log.Debug("Message {0} was found in cache. ", filename );
		  }
		  
		  if (this.index_count % 100 == 1)
		  {
			  MailDirQueryable.log.Debug("Saving cache ({0})", this.index_count);
			  this.SaveCache();
		  }
		}
		
		public static Uri EmailUri (string accountName, string folderName, string uid)
		{
			return new Uri (String.Format ("email://{0}/{1};uid={2}",
						accountName, folderName, uid));
		}

		override protected double RelevancyMultiplier (Hit hit)
		{
			return HalfLifeMultiplierFromProperty (hit, 0.25,
					"fixme:received", "fixme:sentdate");
		}
		
		protected static Stream ReadAppData (string name)
		{
			string path = Path.Combine (Path.Combine (PathFinder.StorageDir, "MailDirIndex"), name);
			return new FileStream (path, System.IO.FileMode.Open, FileAccess.Read, FileShare.Read);
		}

		protected static string ReadAppDataLine (string name)
		{
			Stream stream;

			try {
				stream = ReadAppData (name);
			} catch (FileNotFoundException) {
				return null;
			}

			StreamReader sr = new StreamReader (stream);
			string line = sr.ReadLine ();
			sr.Close ();
			return line;
		}

		protected static Stream WriteAppData (string name)
		{
			string path = Path.Combine (Path.Combine (PathFinder.StorageDir, "MailDirIndex"), name);
			return new FileStream (path, System.IO.FileMode.Create, FileAccess.Write, FileShare.None);
		}

		protected static void WriteAppDataLine (string name, string line)
		{
			if (line == null) {
				string path = Path.Combine (Path.Combine (PathFinder.StorageDir, "MailDirIndex"), name);

				if (File.Exists (path))
					File.Delete (path);

				return;
			}

			StreamWriter sw = new StreamWriter (WriteAppData (name));
			sw.WriteLine (line);
			sw.Close ();
		}
	}
}
