
//
// MailDirIndexableGenerator.cs
//
// Copyright (C) 2005 Noam Rathaus, Aviram Jenik
//
//
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//

using System;
using System.Collections;
using System.IO;
using System.Threading;
using System.Xml;

using Beagle.Util;
using Beagle.Daemon;

using Mozilla = Beagle.Util.Mozilla; // FIXME: We need to try and change this to GMime, if we can
using GConf;
using Mono.Posix;

namespace Beagle.Daemon.MailDirDriver {

	public class MailDirIndexableGenerator : IIndexableGenerator {
		private FileInfo email_info;
		protected bool parsed;
		private MailDirQueryable queryable;

		public MailDirIndexableGenerator (MailDirQueryable queryable, FileInfo file)
		{
			this.queryable = queryable;
			this.email_info = file;
			this.parsed = false;
		}

		public MailDirQueryable Queryable {
			get { return this.queryable; }
		}

		protected bool IsSpamFolder (string name)
		{
			if (name.ToLower () == "spam" || name.ToLower () == "junk" 	|| name.ToLower	() == "filteredspam" ||
					name.ToLower() == "trash") // FIXME: Any other folders?
				return true;
			else
				return false;
		}

		public Indexable GetNextIndexable ()
		{
			string filename = email_info.FullName;
			Indexable indexable = null;
			
			MailDirQueryable.log.Debug("Invoking Mozilla.MessageReader({0})", filename);
			if(File.Exists(filename)) {
	  		Mozilla.MessageReader mr = new Mozilla.MessageReader(filename);
  			Mozilla.Message message = mr.NextMessage;
				MailDirQueryable.log.Debug("message from: {0} to: {1} subject: {2}", message.From, message.To, 
						message.Subject);
				indexable = this.MailDirMessageToIndexable (message);
			}
			this.parsed = true;

			return indexable;
		}


		
		public bool HasNextIndexable ()
		{
			if (this.parsed)
			{
				CrawlFinished();
				return false;
			}
			else
			{
				return true;
			}
		}


		protected bool CrawlNeeded ()
		{
			string timeStr;

			try {
				timeStr = ExtendedAttribute.Get (this.CrawlFile.FullName, "LastCrawl");
			} catch {
				MailDirQueryable.log.Debug ("CrawlNeeded() Unable to get last crawl time on {0}", this.CrawlFile.FullName);
				return true;
			}

			DateTime lastCrawl = StringFu.StringToDateTime (timeStr);

			if (this.CrawlFile.LastWriteTime > lastCrawl)
				return true;
			else
				return false;
		}

		protected void CrawlFinished ()
		{
			try {
				ExtendedAttribute.Set (this.CrawlFile.FullName, "LastCrawl",
						       StringFu.DateTimeToString (DateTime.Now));
			} catch {
				MailDirQueryable.log.Debug ("CrawlFinished() Unable to set last crawl time on {0}", this.CrawlFile.FullName);
			}
		}

		public string StatusName {
			get { return this.CrawlFile.FullName; }
		}

		public override bool Equals (object o)
		{
			MailDirIndexableGenerator generator = o as MailDirIndexableGenerator;

			if (generator == null)
				return false;

			if (Object.ReferenceEquals (this, generator))
				return true;

			if (this.CrawlFile.FullName == generator.CrawlFile.FullName)
				return true;
			else
				return false;
		}

		public override int GetHashCode ()
		{
			MailDirQueryable.log.Debug("this.CrawlFile.FullName.GetHashCode (): {0}", this.CrawlFile.FullName.GetHashCode ());
			return this.CrawlFile.FullName.GetHashCode ();
		}

		private Indexable MailDirMessageToIndexable (Mozilla.Message message)
		{
			MailDirQueryable.log.Debug("MailDirMessageToIndexable()");
			
			MailDirQueryable.log.Debug("UriFu.PathToFileUri({0})", email_info.FullName);
			Uri file_uri = UriFu.PathToFileUri(email_info.FullName);
			
			Indexable indexable = new Indexable (file_uri);

			bool bDateConverted = false;
			try {
				MailDirQueryable.log.Debug("Rfc822DateTime.FromString({0})", message.Date);
				indexable.Timestamp = Rfc822DateTime.FromString(message.Date);
				bDateConverted = true;
				}
			catch {
				MailDirQueryable.log.Debug("Not RFC date");
			}

			if (!bDateConverted) {
				try {
					MailDirQueryable.log.Debug("StringFu.StringToDateTime({0})", message.Date);
					indexable.Timestamp = StringFu.StringToDateTime(message.Date);
					bDateConverted = true;
				}
				catch {
					MailDirQueryable.log.Debug("StringFu.StringToDateTime() failed");
				}
			}
			MailDirQueryable.log.Debug("DateConversion done.");
			indexable.Type = "MailMessage";
			indexable.MimeType = "text/plain"; // FIXME: Some emails aren't text/plain

			indexable.AddProperty (Property.NewKeyword ("dc:title", message.Subject));

			MailDirQueryable.log.Debug("folder: {0}\nsubject: {1}\nto: {2}\nfrom: {3}", email_info.FullName, message.Subject, message.To, message.From);
      indexable.AddProperty (Property.NewKeyword ("fixme:folder", email_info.FullName));
			indexable.AddProperty (Property.NewKeyword ("fixme:subject", message.Subject));
      indexable.AddProperty (Property.NewKeyword ("fixme:to",      message.To));
			indexable.AddProperty (Property.NewKeyword ("fixme:from",    message.From));
//			indexable.AddProperty
			if (email_info.FullName.IndexOf("sent-mail") != -1)
			{
				indexable.AddProperty (Property.NewDate ("fixme:sentdate", indexable.Timestamp));
				indexable.AddProperty (Property.NewFlag ("fixme:isSent"));
				MailDirQueryable.log.Debug("Mail was sent-mail");
			}
			else
			{
				indexable.AddProperty (Property.NewDate ("fixme:received", indexable.Timestamp));
				MailDirQueryable.log.Debug("Mail was received");
			}

			StringReader reader = new StringReader (message.Body);
			indexable.SetTextReader (reader);
			MailDirQueryable.log.Debug("Message indexed done.");
						
			return indexable;
		}

		public void Checkpoint ()
    {
    }

		protected FileInfo CrawlFile {
			get { return this.email_info; }
		}
	}
}
